#ifndef BNETFWD_HH
#define BNETFWD_HH

#include <ext/hash_map>
#include "tcpforward.hh"
#include "udpmx.hh"
#include "bnetproto.hh"

class bnet_forwarder;

class bnet_tcp_proxied_connection : public tcp_forward
{
	friend class bnet_forwarder;
protected:
	struct tcp_state
	{
		_bnet_packet_head pkt_head;
		size_t pkt_offs, subst_offs, subst_len;
		buffer<char> subst_data;
		size_t (bnet_tcp_proxied_connection::* pkt_handler)(virtual_buffer<char>& data);
		size_t packet_bytes_left;

		tcp_state() : pkt_offs(0), pkt_handler(NULL), packet_bytes_left(0)
			{ pkt_head.len = 0; }

		bool new_packet(const char *head);
		bool subst(virtual_buffer<char>& data, buffered_io& io);
	};

	bnet_forwarder *master;

	enum { INIT = 0, VALID, TRANSPARENT } state;
	tcp_state client_state, server_state;
	_bnet_server_challenge challenge;
	bool got_challenge;

	enum { STRUCT, GAMENAME, PADSTRZ, MAPNAME } gamelist_state;
	unsigned cur_game;

	void check_connection_type(forwarding_side *_slave);
	void break_packet(forwarding_side *_slave);
	void slave_read_callback(forwarding_side *_slave);
	void slave_error_callback(forwarding_side *_slave, int err);

	size_t handle_header_from_server_cb(virtual_buffer<char>& part);
	size_t handle_header_from_client_cb(virtual_buffer<char>& part);

	void setup_client_hello();
	size_t process_server_hello(virtual_buffer<char>& data);
	size_t process_server_gameinfo_get_count(virtual_buffer<char>& data);
	size_t process_server_gameinfo(virtual_buffer<char>& data);
public:
	bnet_tcp_proxied_connection(bnet_forwarder *_master) : master(_master), state(INIT), got_challenge(false) {}

	bnet_forwarder *get_master() { return master; }
};

class bnet_tcp_forward_listener : public tcp_forward_listener
{
protected:
	bnet_forwarder *master;

	tcp_forward *create_forward();
	void connected(tcp_forward *fwd);
	void connect_died(int ofd, int cfd, int err);
	bool validate_connection(int ofd, int cfd);
public:
	bnet_tcp_forward_listener(bnet_forwarder *_master, const socket_address& _to)
		: tcp_forward_listener(_to), master(_master) {}
};

class bnet_udp_forwarder;

class bnet_udp_peer : public udp_mx_peer
{
protected:
	void pkt_read();
	void mangle_player_packet(_bnet_udp_packet__player *data);
public:
	bnet_udp_forwarder *via;

	bnet_udp_peer(bnet_udp_forwarder *mx, const socket_address& peer, bnet_udp_forwarder *send_via);
};

class bnet_udp_forwarder : public udp_mx
{
	friend class bnet_udp_peer;
public:
	typedef udp_mx_peer *(bnet_forwarder::*create_func_ptr)(bnet_udp_forwarder *mx, const socket_address& from);
protected:
	socket_address forward_target;
	bnet_forwarder *master;
	create_func_ptr create_func;
	bool accept_all_peers;

	udp_mx_peer *create_peer();
	bool validate_new_peer();
public:
	bnet_udp_forwarder(bnet_forwarder *_master, const socket_address& fwdtarget,
			create_func_ptr createf, bool accept_all = false)
		: master(_master), forward_target(fwdtarget), create_func(createf),
		  accept_all_peers(accept_all) {}
	virtual ~bnet_udp_forwarder();

	void drop_peer(udp_mx_peer *peer);
	const socket_address& get_forward_target() const { return forward_target; }
};

class ip_semaphores
{
protected:
	typedef __gnu_cxx::hash_map<uint32_t, unsigned> ip_set_type;
	ip_set_type set;
public:
	void put(uint32_t ip);
	bool get(uint32_t ip);
};

class bnet_forwarder
{
protected:
	class udp_mx_map : public __gnu_cxx::hash_map<socket_address::ipportpair, bnet_udp_forwarder *>
	{
	public:
		bnet_udp_forwarder::create_func_ptr create_func;
		socket_address bind_address;
		const char *mark_str;
	};

	typedef __gnu_cxx::hash_map<socket_address::ipportpair, socket_address::ipportpair> address_map;

	bnet_tcp_forward_listener tcp_listener;
	bnet_udp_forwarder udp_listener;
	socket_address server;

	ip_semaphores allowed_udp_ips;
	std::set<bnet_tcp_proxied_connection *> clients;
	udp_mx_map *internal_udp;	// gracz zewn. -> port na wewnętrznym if
	udp_mx_map *external_udp;	// klient -> port na zewnętrznym if
	address_map masking_udp;	// port na zewnętrznym -> port klienta
public:
	bnet_forwarder(const socket_address& _listen, const socket_address& _server);
	~bnet_forwarder();

	void new_client(bnet_tcp_proxied_connection *client);
	bool validate_connection(const socket_address& from);
	void connection_error(bnet_tcp_proxied_connection *client, bool server_side, int err);
	bool validate_peer(const socket_address& from);
protected:
	bnet_udp_forwarder *create_udp_masking_port(udp_mx_map& udp_map, const socket_address& addr_masked);
	udp_mx_peer *new_udp_peer(udp_mx_map& udp_map, bnet_udp_forwarder *mx, const socket_address& peer);
	udp_mx_peer *new_udp_client(bnet_udp_forwarder *mx, const socket_address& from);
	udp_mx_peer *new_udp_gamer(bnet_udp_forwarder *mx, const socket_address& from);
public:
	socket_address::ipportpair map_host_for_remote(const socket_address::ipportpair& addr);
	socket_address::ipportpair map_host_for_local(const socket_address::ipportpair& addr);
	socket_address::ipportpair map_host_for_addr(const socket_address::ipportpair& addr, const socket_address::ipportpair& dest);
	void drop_udp_forwarding(bnet_udp_forwarder *fwd);
};

#endif /* !BNETFWD_HH */
