#ifndef BUFFER_HH
#define BUFFER_HH

#include "eventwrap.hh"
#include <iostream>
#include <cstdlib>
#include <algorithm>
#include <queue>
#include <utility>

#define BUFBLOCK        0x1000
#define BUFBLOCKMASK    0x0FFF

template <class el> class buffer;
template <class el> class packet_buffer;

template <class el>
class virtual_buffer
{
	friend class buffer<el>;
	friend class packet_buffer<el>;
protected:
	mutable el *data;
	size_t size, head, tail;
public:
	virtual_buffer() : data(NULL), size(0), head(0), tail(0) {}
	virtual_buffer(el *ptr, size_t len) : data(ptr), size(len), head(0), tail(len) {}
	virtual_buffer(const el *ptr, size_t len) : data(ptr), size(len), head(0), tail(len) {}
	virtual_buffer(el *start, el *end) : data(start), size(end - start), head(0), tail(end - start) {}
	virtual ~virtual_buffer() {}

	void clear()
	{
		tail = head = 0;
	}

	bool is_empty() const
	{
		return tail == head;
	}

	size_t count_elements_left() const
	{
		return tail - head;
	}

	void advance(size_t howmany)
	{
		head += howmany;
		if (head > tail)
			head = tail;
	}

	el *get_head_ptr() const { return data + head; }
	el *get_tail_ptr() const { return data + tail; }

private:
	void my_copy(const virtual_buffer<el>& src)
	{
		data = src.data + src.head;
		head = 0;
		size = tail = src.tail - src.head;
	}

	void my_copy(const virtual_buffer<el>& src, const size_t n)
	{
		my_copy(src);
		if (size > n)
			size = tail = n;
	}

public:
	virtual_buffer(const virtual_buffer<el>& src)
	{
		my_copy(src);
	}

	virtual_buffer(const virtual_buffer<el>& src, const size_t n)
	{
		my_copy(src, n);
	}

	virtual virtual_buffer<el>& copy(const virtual_buffer<el>& src)
	{
		my_copy(src);
		return *this;
	}

	virtual virtual_buffer<el>& copy(const virtual_buffer<el>& src, const size_t n)
	{
		my_copy(src, n);
		return *this;
	}

	virtual_buffer<el>& operator=(const virtual_buffer<el>& src)
	{
		return copy(src);
	}
};

template <class el>
class buffer : public virtual_buffer<el>
{
public:
	buffer() : virtual_buffer<el>() {}

	void free()
	{
		if (this->data) {
			std::free(this->data);
			this->data = NULL;
			this->size = this->head = this->tail = 0;
		}
	}

	~buffer() { free(); }

	void realloc(size_t nsize)
	{
		nsize += BUFBLOCKMASK;
		nsize &= ~BUFBLOCKMASK;

		if (this->size != nsize)
			this->data = (el *)std::realloc(this->data, (this->size = nsize) * sizeof(el));
	}

	void clean()
	{
		if (!this->head)
			return;
		std::copy_backward(this->data + this->head, this->data + this->tail, this->data + this->tail - this->head);
		this->tail -= this->head;
		this->head = 0;
	}

	void expand(size_t more)
	{
		if (this->tail + more > this->size)
			realloc(this->tail + more);
	}

	void shrink(size_t less)
	{
		if (less > this->tail - this->head)
			less = this->tail - this->head;

		this->tail -= less;
	}

	void append(const el *const ptr, const size_t len)
	{
		expand(len);
		std::copy(ptr, ptr + len, this->get_tail_ptr());
		this->tail += len;
	}

	void append(const el *const ptr);

	void append(const virtual_buffer<el>& src)
	{
		append(src.get_head_ptr(), src.count_elements_left());
	}

	void append_and_move(virtual_buffer<el>& src, size_t copylen, size_t skiplen = 0)
	{
		if (copylen > src.count_elements_left())
			copylen = src.count_elements_left();
		append(src.get_head_ptr(), copylen);
		src.advance(copylen + skiplen);
	}

	el *append_space(size_t howmany)
	{
		expand(howmany);
		el *p = this->data + this->tail;
		this->tail += howmany;
		return p;
	}

private:
	void my_copy(const virtual_buffer<el>& src)
	{
		this->head = 0;
		realloc(this->tail = src.count_elements_left());
		if (src.data + src.head > this->data)
			std::copy(src.data + src.head, src.data + src.tail, this->data);
		else	std::copy_backward(src.data + src.head, src.data + src.tail, this->data);
	}

	void my_copy(const virtual_buffer<el>& src, const size_t n)
	{
		this->head = 0;
		this->tail = src.count_elements_left();
		if (this->tail > n)
			this->tail = n;
		realloc(this->tail);
		if (src.data + src.head > this->data)
			std::copy(src.data + src.head, src.data + src.head + this->tail, this->data);
		else	std::copy_backward(src.data + src.head, src.data + src.head + this->tail, this->data);
	}

public:
	buffer(const virtual_buffer<el>& src)
	{
		my_copy(src);
	}
	buffer(const virtual_buffer<el>& src, const size_t n)
	{
		my_copy(src, n);
	}

	virtual_buffer<el>& copy(const virtual_buffer<el>& src)
	{
		my_copy(src);
		return *this;
	}

	virtual_buffer<el>& copy(const virtual_buffer<el>& src, const size_t n)
	{
		my_copy(src, n);
		return *this;
	}

	bool readline(virtual_buffer<el>& linedst, const virtual_buffer<el>& rl_ending);
};

template <class el>
class packet_buffer : protected buffer<el>
{
protected:
	typedef std::pair<size_t, size_t> q_entry;
	std::queue<q_entry> q;
	size_t prepended_tail;
	bool prepending;
public:
	packet_buffer() : prepended_tail(0), prepending(false) {}

	void push(const virtual_buffer<el>& pkt);
	void pop();
	virtual_buffer<el> get_next() const;
	bool is_empty() const { return q.empty(); }

	void clean(size_t want_free = 0);

	friend std::ostream& operator<<(std::ostream& os, packet_buffer<el>& pb)
	{
		os << pb.prepended_tail << ',' << pb.head << '-' << pb.tail << '/' << pb.size;
		return os << (pb.prepending ? "+" : "-") << ' ' << pb.q.size();
	}
};

template <>
void buffer<char>::append(const char *const ptr);

template <>
bool buffer<char>::readline(virtual_buffer<char>& alinedst, const virtual_buffer<char>& rl_ending);

template <class el>
void packet_buffer<el>::push(const virtual_buffer<el>& pkt)
{
	if (!prepending && this->size - this->tail < pkt.count_elements_left())
		prepending = true;
	if (prepending && this->head - prepended_tail < pkt.count_elements_left())
		clean(pkt.count_elements_left());

	size_t& where = prepending ? prepended_tail : this->tail;
	std::copy(pkt.get_head_ptr(), pkt.get_tail_ptr(), this->data + where);

	q_entry entry;
	entry.first = where;
	entry.second = where += pkt.count_elements_left();
	q.push(entry);
}

template <class el>
void packet_buffer<el>::pop()
{
	q_entry entry = q.front();
	q.pop();

	if (prepending && this->head == this->tail) {
		this->tail = prepended_tail;
		prepended_tail = this->head = 0;
		prepending = false;
	}

	if (entry.first != this->head)
		abort();

	this->head = entry.second;
}

template <class el>
virtual_buffer<el> packet_buffer<el>::get_next() const
{
	q_entry entry = q.front();
	virtual_buffer<el> buf;

	buf.data = this->data;
	buf.size = this->size;
	buf.head = entry.first;
	buf.tail = entry.second;

	return buf;
}

template <class el>
void packet_buffer<el>::clean(size_t want_free)
{
	if (prepending) {
		size_t need_free = this->tail + prepended_tail - this->size;
		if (want_free < need_free)
			want_free = need_free;

		this->expand(want_free);
		std::copy(this->data, this->data + prepended_tail, this->data + this->tail);
		this->tail += prepended_tail;
		prepended_tail = 0;
		prepending = false;
	} else {
		size_t have_free = this->size - this->tail + this->head;
		if (want_free > have_free)
			expand(want_free - this->head);
	}

	buffer<el>::clean();
}

#endif /* !BUFFER_HH */
